"""
Copyright (c) 2021 Christopher Harrison

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
Public License for more details.

You should have received a copy of the GNU General Public License along
with this program. If not, see https://www.gnu.org/licenses/
"""

from __future__ import annotations

from abc import ABCMeta, abstractmethod


class Eq(metaclass=ABCMeta):
    """ Abstract base class for equality checking """
    @abstractmethod
    def __eq__(self, rhs:Eq) -> bool:
        """ Equality """

    def __neq__(self, rhs:Eq) -> bool:
        """ Inequality """
        return not self == rhs
